use std::collections::HashMap;
use std::sync::mpsc;
//use std::sync::mpsc::TryRecvError;

use pool::Pool;
use super::{Bus, Packet, QueryPacket, Query, Response, Mode};

pub struct Client<T> where T: 'static + Clone + Send {
    pub packet:Option<mpsc::Sender<Packet<T>>>,
    pub query:mpsc::Sender<Response<T>>,
    pub signal:Option<mpsc::Sender<()>>,
}

pub struct Manager<T> where T: 'static + Clone + Send {
    pub receiver:mpsc::Receiver<Packet<T>>,
    pub sender:mpsc::Sender<Packet<T>>,
    pub senders:Pool<Client<T>>,
    pub query:mpsc::Receiver<QueryPacket>,
    pub query_sender:mpsc::Sender<QueryPacket>,
    pub response:mpsc::Receiver<Response<T>>,
    pub response_sender:mpsc::Sender<Response<T>>,
    pub ports:HashMap<u32, u32>,
}
impl<T: 'static + Clone + Send> Manager<T> {
    pub fn new() -> Self
    {
        let (p_sender, p_receiver) = mpsc::channel::<Packet<T>>();
        let (q_sender, q_receiver) = mpsc::channel::<QueryPacket>();
        let (r_sender, r_receiver) = mpsc::channel::<Response<T>>();

        Self {
            receiver:p_receiver,
            sender:p_sender,
            senders:Pool::<Client<T>>::new(),
            query:q_receiver,
            query_sender:q_sender,
            response:r_receiver,
            response_sender:r_sender,
            ports:HashMap::<u32,u32>::new(),
        }
    }
}

pub fn manager_thread<T>(manager:&mut Manager<T>) where T: 'static + Clone + Send
{
    // listen for queries until all clients disconnect
    while manager.senders.size() > 0 {
        let query = manager.query.recv();

        match query {
            Ok(packet) => {
                // process query packet
                match packet.data {
                    Query::Notify => {
                        // receive and process all packets in the queue
                        loop {
                            let recv = manager.receiver.try_recv();
                            match recv {
                                Ok(packet) => {
                                    // forward packet to receiver if exists
                                    if packet.to != 0 {
                                        let dest = manager.senders.get(packet.to as usize);
                                        match dest {
                                            Some(sender) => {
                                                match &sender.packet {
                                                    Some(transmitter) => { transmitter.send(Clone::clone(&packet)).ok(); }
                                                    None => { }
                                                }
                                                match &sender.signal {
                                                    Some(signal) => { signal.send(()).ok(); }
                                                    None => { }
                                                }
                                            }
                                            None => { }
                                        }
                                    }
                                    else {
                                        for to in manager.senders.list() {
                                            if to != packet.from as usize {
                                                let dest = manager.senders.get(to);
                                                match dest {
                                                    Some(sender) => {
                                                        match &sender.packet {
                                                            Some(transmitter) => { transmitter.send(Clone::clone(&packet)).ok(); }
                                                            None => { }
                                                        }
                                                        match &sender.signal {
                                                            Some(signal) => { signal.send(()).ok(); }
                                                            None => { }
                                                        }
                                                    }
                                                    None => { }
                                                }
                                            }
                                        }
                                    }
                                }
                                Err(_) => { break; }
                            }
                        }
                    }
                    
                    Query::Connect(mode) => {
                        // return a new bus endpoint

                        let transmitter :bool;
                        let receiver :bool;
                        match mode {
                            Mode::Transmitter => {
                                transmitter = true;
                                receiver = false;
                            }
                            Mode::Transceiver => {
                                transmitter = true;
                                receiver = true;
                            }
                        }

                        let sender = manager.sender.clone();
                        let (c_sender, c_receiver) = mpsc::channel::<Packet<T>>();
                        let cq_sender = manager.query_sender.clone();
                        let (cr_sender, cr_receiver) = mpsc::channel::<Response<T>>();

                        let id = manager.senders.add(Client {
                            packet:if receiver { Some(c_sender) } else { None },
                            query:cr_sender,
                            signal:None,
                        });

                        let bus = Bus::<T> {
                            sender:if transmitter { Some(sender) } else { None },
                            receiver:if receiver { Some(c_receiver) } else { None },
                            query:cq_sender,
                            response:cr_receiver,
                            id:id as u32,
                        };

                        let client = manager.senders.get(packet.from as usize);
                        match client {
                            Some(client) => { client.query.send(Response::Connect(bus)).ok(); }
                            _ => { }
                        }
                    }

                    Query::Signal(signal) => {
                        let client = manager.senders.get_mut(packet.from as usize);
                        match client {
                            Some(client) => {
                                client.signal = signal;
                                client.query.send(Response::Ok).ok();
                            }
                            _ => { }
                        }
                    }

                    Query::List => {
                        // return a list of valid clients
                        let client = manager.senders.get(packet.from as usize);
                        match client {
                            Some(client) => { client.query.send(Response::List(manager.senders.list())).ok(); }
                            _ => { }
                        }
                    }

                    Query::SetPort(id) => {
                        // return a list of valid clients
                        let client = manager.senders.get(packet.from as usize);
                        match client {
                            Some(client) => {
                                manager.ports.insert(id, packet.from);
                                client.query.send(Response::Ok).ok();
                            }
                            _ => { }
                        }
                    }

                    Query::GetPort(id) => {
                        // return 
                        let client = manager.senders.get(packet.from as usize);
                        match client {
                            Some(client) => {
                                let port = match manager.ports.get(&id) {
                                    Some(id) => Some(*id),
                                    None => None
                                };
                                client.query.send(Response::GetPort(port)).ok();
                            }
                            _ => { }
                        }
                    }
                }
            }
            Err(_) => {

            }
        }
    }
}
